/*
 * Mandel by Rob Swiston and Chuck Simmons
 * Copyright (C) 1989, 1990 by Rob Swiston and Chuck Simmons
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 1 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Here we have the client side of the mandelbrot program.  Primarily,
 * we create connections to servers and farm out portions of the
 * computation to the servers.  Periodically, we give control back
 * to our caller, telling our caller whether or not we are completely
 * done filling a window.
 */

/*
 * Initialize the client.  Here, we want to establish connections to
 * our servers.  We return a context area.
 */

#include <sys/types.h>
#include <sys/timeb.h>
#ifndef DST_NONE
# include <sys/time.h>
#endif

#include <stdio.h>
#include "mandel.h"
#include "gui.h"
#include "net.h"
#include "client.h"

#define HOSTLEN 40

/*
 * Specify the number of lines we want to have outstanding for
 * each server.  Setting this too high will lead to long delays
 * when drawing a new window.  The reason for this is that when
 * the window is nearly completely drawn, we will send each line
 * to many servers to make sure we get a response back.  Setting
 * this too low will lead to a very minor performance degradation
 * (in theory) because servers will have nothing to do while waiting
 * for a line to arrive.
 */
#define NLINES 2

#define bswap(x) ((((x) << 8) & 0xff00) | (((x) >> 8) & 0xff))

typedef struct line_id {
	struct line_id *next; /* pointer to next line in queue */
	int winid; /* identifier of window for this line */
	int line; /* number of line being worked on */
} line_id_t;

typedef struct connection {
	int sock; /* socket descriptor */
	char host[HOSTLEN]; /* name of host */
	line_id_t *q_head; /* head of q of lines */
	line_id_t *q_tail; /* tail of queue */
	int flops; /* estimated flops performed */
} connection_t;

typedef struct client_info {
	color_x_t *buf; /* buffer colors for a line */

	int ncon; /* number of connections */
	connection_t *con; /* connections to servers */
	fd_set open_set; /* list of open sockets */
	int max_fd; /* maxium socket number */

	int winid; /* identifier for window we are drawing */
	int line; /* next line to work on */
	int ndrawn_lines; /* count of drawn lines */
	char *drawn_line; /* list of lines that have been drawn */

	int flops; /* estimated floating point operations performed */
	time_t start_time; /* real time at which we started window */
} client_info_t;

/*
 * Operations on the line queue.
 */

/* Allocate and free a line. */

#define line_alloc() (line_id_t *) malloc (sizeof (line_id_t))
#define line_free(p) free ((char *) p)

/*
 * Add a line to the end of a connection's queue.  Return the
 * added line.
 */

line_id_t *
line_push (conp)
connection_t *conp;
{
	line_id_t *p;

	p = line_alloc ();
	if (!p) return NULL;

	if (conp->q_tail) conp->q_tail->next = p;
	else conp->q_head = p;

	conp->q_tail = p;
	p->next = NULL;
	return p;
}

/*
 * Pop a line off the queue.
 */

void
line_pop (conp)
connection_t *conp;
{
	line_id_t *p;

	p = conp->q_head;

	conp->q_head = p->next;
	if (conp->q_head == NULL) conp->q_tail = NULL;
	line_free (p);
}

/*
 * Return TRUE if a connection has fewer than NLINES lines outstanding.
 */

int
client_has_room (conp)
connection_t *conp;
{
	line_id_t *p;
	int i;

	for (i = 0, p = conp->q_head; i < NLINES && p; i++, p = p->next);

	return i < NLINES;
}

/* Return the y-coordinate of a line. */

double
line_y_coord (winp, line)
char *winp;
int line;
{
	return win_ymin(winp) + (line * win_rheight(winp)) / win_pheight(winp);
}

/* Return the line of a y-coordinate. */

int
y_line_coord (winp, y)
char *winp;
double y;
{
	double t;
	t = win_pheight (winp) * (y - win_ymin (winp)) / win_rheight (winp);
	return (int) (t + 0.5); /* round to nearest int */
}

/*
 * Initialize the client.  We allocate a context area, and we
 * establish connections to each server.
 */

char *
client_init (winp)
char *winp;
{
	client_info_t *clientp;
	FILE *hostf;
	char name[HOSTLEN];
	int port;
	int buflen;
	int ncons;
	connection_t *con;

	clientp = (client_info_t *) malloc (sizeof (*clientp));
	if (!clientp) {
		(void) fprintf (stderr, "client_init: Out of storage.\n");
		exit (1);
	}

	/*
	 * Allocate storage to hold the colors of 1 line and
	 * to hold the list of drawn lines.  We allocate the
	 * maximum amount of storage that we might need for a window.
	 * Note that our buffer pretty much really needs to be this
	 * large.  Consider the case where we send out requests for
	 * a long line, and then the window shrinks.  The servers
	 * well send back really large requests which we would like
	 * to have fit in our buffer.
 	 */
	buflen = sizeof (clientp->buf[0]) * win_max_width (winp);
	clientp->buf = (color_x_t *) malloc (buflen);

	buflen = sizeof (clientp->drawn_line[0]) * win_max_height (winp);
	clientp->drawn_line = (char *) malloc (buflen);

	if (!clientp->buf || !clientp->drawn_line) {
		(void) fprintf (stderr, "client_init: Out of storage.\n");
		exit (1);
	}

	clientp->winid = 0;
	FD_ZERO (&clientp->open_set);
	clientp->max_fd = -1;
	clientp->ncon = 0;

	/* Create connections to the servers. */
	hostf = fopen (HOSTFILE, "r");
	if (!hostf) {
		perror (HOSTFILE);
		exit (1);
	}

	ncons = client_lines (hostf);
	buflen = sizeof (clientp->con[0]) * ncons;
	clientp->con = (connection_t *) malloc (buflen);
	if (!clientp->con) {
		(void) fprintf (stderr, "client_init: Out of storage.\n");
		exit (1);
	}

	while (client_scan_line (hostf, name, &port, &ncons)) {
		(void) fprintf (stderr,
			"Connecting to %s at port %d (%d connections)\n",
			name, port, ncons);

		while (ncons > 0) { /* create 'ncons' connections to this obj */
			con = &(clientp->con[clientp->ncon]);

			(void) strcpy (con->host, name);
			con->q_head = con->q_tail = NULL;
			con->sock = net_copen (con->host, port);
			if (con->sock < 0) {
				con->sock = net_copen (con->host, bswap (port));
			}
			if  (con->sock < 0) {
				(void) fprintf (stderr,
					"Cannot open socket to %s: ",
					con->host);
				perror ("");
				break; /* give up on this host */
			}
			else {
				clientp->ncon += 1;
				FD_SET (con->sock, &clientp->open_set);
				if (con->sock > clientp->max_fd) {
					clientp->max_fd = con->sock;
				}
			}

			ncons -= 1;
		}
	}

	if (clientp->ncon == 0) {
		(void) fprintf (stderr, "client_init: No connections.\n");
		exit (1);
	}

	(void) fclose (hostf);
	return (char *) clientp;
}

/*
 * Read a line from the host file.
 */

int
client_scan_line (f, name, portp, nconsp)
FILE *f;
char *name;
int *portp;
int *nconsp;
{
	char buf[256];
	int status;

	if (fgets (buf, HOSTLEN, f) == NULL) {
		return 0; /* eof */
	}
	status = sscanf (buf, "%s %d %d", name, portp, nconsp);
	if (status != 3) {
		(void) fprintf (stderr,
			"client_scan_line: syntax error in line '%s'.\n",
			buf);
		(void) fprintf (stderr,
			"    format is 'name port_num num_connections'.\n");
		return 0;
	}
	return 1;
}

/*
 * Count the number of connections we need to make.
 */

int
client_lines (f)
FILE *f;
{
	char buf[256];
	char name[HOSTLEN];
	int port, ncons;
	int total;

	total = 0;
	while (client_scan_line (f, name, &port, &ncons)) {
		total += ncons;
	}

	rewind (f);
	return total;
}

/*
 * Reset the client.  Here, we farm out an initial line of the
 * window to each server after resetting ourselves to begin
 * drawing the first line of the window.
 */

int
client_reset (clientp, winp)
client_info_t *clientp;
char *winp;
{
	void client_send ();

	int i, j;

	bzero (clientp->drawn_line,
		sizeof (clientp->drawn_line[0]) * win_pheight (winp));

	clientp->winid += 1; /* change window id */
	clientp->line = -1; /* reset starting line */
	clientp->ndrawn_lines = 0; /* nothing drawn yet */
	clientp->flops = 0; /* no floating point ops done yet */
	clientp->start_time = time ((time_t *) 0);

	for (i = 0; i < clientp->ncon; i++) {
		clientp->con[i].flops = 0;
	}

	/*
	 * Send NLINES lines to each connection.  We loop on
	 * the connections in the inner loop so that consecutive
	 * lines will be sent to different processors.  We only send
	 * lines to those processors who are in need of additional lines.
	 */
	for (j = 0; j < NLINES; j++)
	for (i = 0; i < clientp->ncon; i++) {
		if (client_has_room (&(clientp->con[i]))) {
			client_send (clientp, &(clientp->con[i]), winp);
		}
	}

	return client_resume (clientp, winp);
}

/*
 * Close down a server connection after an error occurs.
 */

void
client_close (clientp, conp)
client_info_t *clientp;
connection_t *conp;
{
	int i;

	i = conp - &(clientp->con[0]);
	FD_CLR (conp->sock, &clientp->open_set);
	close (conp->sock);
	while (conp->q_head) line_pop (conp); /* free some storage */

	if (i != clientp->ncon-1) { /* compress list of clients */
		clientp->con[i] = clientp->con[clientp->ncon-1];
	}
	clientp->ncon -= 1;

	if (clientp->ncon == 0) {
		(void) fprintf (stderr,
			"client_close: All connections died.\n");
		exit (1);
	}
}

/*
 * Send a line out over the network, updating our client structure.
 */

void
client_send (clientp, conp, winp)
client_info_t *clientp;
connection_t *conp;
char *winp;
{
	extern int arg_iterations;

	int status;
	line_id_t *p;

	if (clientp->ndrawn_lines >= win_pheight (winp)) return;

	/* increment client's line index to next undrawn line */
	do {
		if (clientp->line == win_pheight (winp) - 1) {
			clientp->line = 0;
		}
		else clientp->line += 1;
	} while (clientp->drawn_line[clientp->line]);

	p = line_push (conp);
	if (!p) return; /* don't send anything if we don't have the storage */

	p->winid = clientp->winid;
	p->line = clientp->line;

	status = net_csend (conp->sock, win_xmin (winp),
			line_y_coord (winp, clientp->line),
			win_xmax (winp), win_pwidth (winp), arg_iterations);

	if (status < 0) {
		(void) fprintf (stderr,
			"client_send: net_csend to %s failed: ",
			conp->host);
		perror ("");
		client_close (clientp, conp);
	}
}

/*
 * Receive a buffer off the network, and draw it.  Also, send another
 * line.
 */

void
client_recv (clientp, conp, winp)
client_info_t *clientp;
connection_t *conp;
char *winp;
{
	int buflen;
	int winid;
	int line, line2;
	int last_x, i, flops;
	int draw_it;
	double y;

	buflen = net_crecv (conp->sock, clientp->buf);
	if (buflen < 0) {
		(void) fprintf (stderr,
			"client_recv: net_crecv from %s failed: ",
			conp->host);
		perror ("");
		client_close (clientp, conp);
	}
	else {
		winid = conp->q_head->winid;
		line = conp->q_head->line;
		line_pop (conp);

		/* do nothing if this line was for a previous window */
		draw_it = FALSE; /* assume this line is boring */
		line2 = -1; /* no symmetry found yet */
		if (winid == clientp->winid && !clientp->drawn_line[line]) {
			/* mark this line as drawn */
			clientp->drawn_line[line] = TRUE;
			clientp->ndrawn_lines += 1;
			draw_it = TRUE; /* remember to draw the line */
		}

		/*
		 * Check for symmetry about the X-Axis.  If we have symmetry,
		 * get the line number of the symmetrical line.  Do this here
		 * so that we will mark the line of symmetry as being drawn
		 * before sending out the next line to compute.
		 */
		if (draw_it ) {
			/* get y-coordinate of symmetrical line */
			y = -line_y_coord (winp, line);
			if (y >= win_ymin (winp) && y <= win_ymax (winp)) {
				line2 = y_line_coord (winp, y);

				if (line2 >= 0 && line2 < win_pheight (winp) &&
					!clientp->drawn_line[line2])
				{
					clientp->drawn_line[line2] = TRUE;
					clientp->ndrawn_lines += 1;
				}
				else line2 = -1; /* don't draw line 2 */
			}
		}

		/* send a new line out quickly */
		client_send (clientp, conp, winp);
		if (!draw_it) return;

		/* Compute flops used to calculate this line. */
		flops = 0;
		last_x = 0;
		for (i = 0; i < buflen; i++) {
			flops += clientp->buf[i].color
					* (clientp->buf[i].x - last_x + 1);
			last_x = clientp->buf[i].x + 1;
		}
		clientp->flops += flops;
		conp->flops += flops;

		/* draw the line on the screen */
		win_render_line (winp, clientp->buf, buflen, line);

		if (line2 != -1) { /* draw the symmetrical line? */
			win_rerender_line (winp, line2);
		}
	}
}

/*
 * Resume a client wherever we left off.  We wait for a server
 * to report that she has finished computing a line.  We then
 * draw the line and farm out a new line, if we have one available.
 *
 * When we receive a line from a client, or when a timeout expires,
 * we will give control to our caller so that user input can be processed.
 */

int
client_resume (clientp, winp)
client_info_t *clientp;
char *winp;
{
	fd_set read_set;
	struct timeval tv;
	int nfound;
	int max_fd;
	int buflen;
	int i;

	tv.tv_sec = 1; /* init timeout */
	tv.tv_usec = 0;

	read_set = clientp->open_set;
#if 0
(void) fprintf (stderr, "before read_set=0x%x, max_fd=%d\n",
		((int *)(&read_set))[0], clientp->max_fd);
#endif
	nfound = select (clientp->max_fd+1, &read_set, (fd_set *)NULL,
			(fd_set *)NULL, &tv);
#if 0
(void) fprintf (stderr, "after read_set=0x%x, max_fd=%d\n",
		((int *)(&read_set))[0], clientp->max_fd);
#endif

	if (nfound < 0) {
		perror ("client_resume: select failed");
		exit (1);
	}

	/* scan for events */
	for (i = 0; i < clientp->ncon && nfound; i++) {
		if (!FD_ISSET (clientp->con[i].sock, &read_set)) {
			continue;
		}

		nfound -= 1;
		client_recv (clientp, &(clientp->con[i]), winp);
	}

	if (clientp->ndrawn_lines < win_pheight (winp)) {
		return TRUE; /* resume later */
	}
	else {
		time_t elapsed;

		/*
		 * Report the flops rating.
		 */
		elapsed = time ((time_t *)0) - clientp->start_time;
		if (elapsed == 0) elapsed = 1; /* a little white lie */

		(void) printf ("Megaflops: %10.2f\n",
			(clientp->flops  * 9.0) / (elapsed * 1000000.0));

		for (i = 0; i < clientp->ncon; i++) {
			(void) printf ("%16s: %10.2f\n",
				clientp->con[i].host,
				(clientp->con[i].flops * 9.0)
					/ (elapsed * 1000000.0));
		}
		(void) fflush (stdout);


		return FALSE; /* tell caller we need a new window */
	}
}
