/*
 * Mandel by Rob Swiston and Chuck Simmons
 * Copyright (C) 1989, 1990 by Rob Swiston and Chuck Simmons
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 1 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * This file contains the code to perform mandelbrot calculations.
 * We are given a description of the line segment on the real plane
 * and the number of pixels that will be used to display that line
 * segment.  For each pixel, we compute a color to be used for drawing
 * that segment.  The resulting pixels are run-length encoded and
 * stored into a passed buffer.
 */

#include "mandel.h"

/* Return the number of elements we put in the buffer. */

int
mandel_line (x_start, y_start, x_end, npixels, max_color, buf)
double x_start, y_start; /* starting co-ordinates of line segment */
double x_end; /* ending co-ordinate of line segment */
int npixels; /* number of pixels to compute */
int max_color; /* maximum color we should use */
color_x_t *buf; /* output -- buffer of 'npixels' elements that we fill */
{
	double x, x_inc;
	int buflen;
	int i, color, old_color;

	buflen = 0; /* nothing in buffer yet */
	x = x_start;
	x_inc = (x_end - x_start) / (double) npixels;
	for (i = 0; i < npixels; i++) {
		color = get_color (x, y_start, max_color);

		if (i == 0) old_color = color;
		else if (color != old_color) {
			buf[buflen].x = i-1;
			buf[buflen].color = old_color;
			buflen += 1;
			old_color = color;
		}

		x += x_inc;
	}

	/* save the last color */
	buf[buflen].x = i-1;
	buf[buflen].color = old_color;
	buflen += 1;

	return buflen;
}

/*
 * Given a location on a plane, perform up to N iterations on that point
 * in order to figure out whether or not the point is in the Mandelbrot
 * Set.
 *
 * We return a value in the range 0..ncolors.  If 'ncolors' is returned,
 * the point is probably in the set.  Lower values indicate how "close" the
 * point is to the set.  The lower the value, the farther away the point is.
 *
 * We loop computing "z = z*z + c", where 'c' is the initial complex
 * number 'x + i*y', and 'z' is a complex variable.  'c' is in the mandelbrot
 * set if, as N goes to infinity, the values assigned to 'z' converge.
 * Basically, if |z| > 4, then we know that the value won't converge.
 *
 * This version is written so as to encourage compilers to overlap
 * instructions and perform other optimizations. 
 */

#ifndef NCUBE
int
get_color (cx, cy, ncolors)
register double cx, cy;
register int ncolors;
{
	register int i;
	register double zx, zy, xx, xy, yy, t, xy2, four;

	zx = cx;
	zy = cy;
	four = 4.0;
	i = 0;

	for (;;) {
		xy = zx * zy;
		xx = zx * zx;
		xy2 = xy + xy;
		yy = zy * zy;

		zy = xy2 + cy;
		t = xx + yy;
		zx = xx - yy + cx;

		if (t > four) break;
		i += 1;
		if (i == ncolors) break;
	}

	return i;
}
#endif
