/*
 * Mandel by Rob Swiston and Chuck Simmons
 * Copyright (C) 1989, 1990 by Rob Swiston and Chuck Simmons
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 1 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * This file contains all the routines used to implement the
 * graphical user interface.  The complete user interface is
 * implemented in this file.  If you want to change windowing
 * systems, you need to rewrite this file.
 *
 * This file interfaces with the client server.  From the user
 * interface, we can call routines to initialize the client,
 * reset the client to operate with a new window, and resume the
 * client to continue operating with an existing window.
 *
 * The client can in turn call some of the routines in this file
 * to draw on the current window.
 */

#include <stdio.h>
#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>

#ifdef apollo
# include </sys5.3/usr/include/math.h>
#else
# include <math.h>
#endif

#include "mandel.h"
#include "gui.h"
#include "client.h"

/* Don't include 'malloc.h' since it doesn't exist on BSD systems. */
char *malloc ();
void  free ();

/* Let's not take a chance with 'string(s).h'. */
char *strchr ();

/* pointers to graphics argument strings */
char  *arg_display = 0;
char  *arg_geometry = 0;
int    arg_use_root = 0;
int    arg_iterations = 256;
double arg_x_center = 0.0;
double arg_y_center = 0.0;
double arg_pix_size = 4.0 / HPIXELS;
int    arg_mud_num = 3;
double arg_mud_max = 0.60;
char  *arg_colorfile = 0;
int    arg_show_time = 0;

char *prog_name;

typedef struct loc {
    int x, y; /* coordinates of a pixel */
} loc_t;

/* Define a context area for holding information about a display area. */
typedef struct win_info {
	Display *dsp;
	Window win;
	union { Pixmap p; Window w; } pix; /* holds the image */
	GC gc; /* graphics context */
	XImage *image; /* imaging buffer */
	int color_len; /* last pixel in colormap we can use */
	unsigned long *pixels;
	Cursor busy; /* cursor to use when computing */
	Cursor nbusy; /* cursor to use when not computing */
	int black, white; /* black and white pixels */

	/* describe the imaging area */
	int hpixels, vpixels; /* size of display area in pixels */
	double x, y; /* center of display area */
	double pix_size; /* size of a pixel */

	/* The following is used when drawing on the root window. */
#define SAVE_LIST_LEN 1
	loc_t black_list[SAVE_LIST_LEN];
	loc_t color_list[SAVE_LIST_LEN];
	int black_num; /* number of items in black_list */
	int color_num; /* number of items in color_list */
} win_info_t;

/*
 * The following routines exist to hide information from the client
 * routine.
 */

/* width and height of window in pixels */
#define W_PWIDTH(w)  ((w)->hpixels)
#define W_PHEIGHT(w) ((w)->vpixels)

/* width and height in real numbers */
#define W_RWIDTH(w)  ((w)->hpixels * (w)->pix_size)
#define W_RHEIGHT(w) ((w)->vpixels * (w)->pix_size)

/* co-ordinates of window edges */
#define W_XMIN(w)    ((w)->x - W_RWIDTH(w) / 2)
#define W_XMAX(w)    ((w)->x + W_RWIDTH(w) / 2)
#define W_YMIN(w)    ((w)->y - W_RHEIGHT(w) / 2)
#define W_YMAX(w)    ((w)->y + W_RHEIGHT(w) / 2)

int win_pwidth  (winp) win_info_t *winp; { return W_PWIDTH (winp); }
int win_pheight (winp) win_info_t *winp; { return W_PHEIGHT (winp); }

/* 'real' co-ordinates of upper-left and lower-right of window */
double win_xmin (winp) win_info_t *winp; { return W_XMIN (winp); }
double win_xmax (winp) win_info_t *winp; { return W_XMAX (winp); }
double win_ymin (winp) win_info_t *winp; { return W_YMIN (winp); }
double win_ymax (winp) win_info_t *winp; { return W_YMAX (winp); }

/* 'real' width and height */
double win_rwidth  (winp) win_info_t *winp; { return W_RWIDTH (winp); }
double win_rheight (winp) win_info_t *winp; { return W_RHEIGHT (winp); }

/* maximum width and height */
int
win_max_width (winp)
win_info_t *winp;
{
	return DisplayWidth (winp->dsp, DefaultScreen (winp->dsp));
}

int
win_max_height (winp)
win_info_t *winp;
{
	return DisplayHeight (winp->dsp, DefaultScreen (winp->dsp));
}

#define OPTIONS "c:d:g:i:m:n:p:rs:x:y:"

void
usage (err)
char *err;
{
	if (err) fprintf (stderr, "%s: %s\n", prog_name, err);
	fprintf (stderr, "Usage: %s [options...]\n", prog_name);
	fprintf (stderr, "where options include:\n");
	fprintf (stderr, "    -c colorfile        file of colors to use\n");
	fprintf (stderr, "    -d host:display     X server to use\n");
	fprintf (stderr, "    -g geometry         size of window\n");
	fprintf (stderr, "    -i count            iteration count\n");
	fprintf (stderr, "    -m mud_max          max val for muddy colors\n");
	fprintf (stderr, "    -n mud_num          # of muddy colors per hue\n");
	fprintf (stderr, "    -p pixel_size       size of a pixel\n");
	fprintf (stderr, "    -r                  draw on root window\n");
	fprintf (stderr, "    -s show_time        show initial color map\n");
	fprintf (stderr, "    -x x                x coord of center\n");
	fprintf (stderr, "    -y y                y coord of center\n");
	exit (1);
}

void
parse_args (argc, argv)
int argc;
char **argv;
{
    extern char *optarg;
    extern int optind;

    int c;

    prog_name = argv[0];

    while ((c = getopt (argc, argv, OPTIONS)) != -1) {
	switch (c) {
	case 'c': arg_colorfile = optarg; break;
	case 'd': arg_display   = optarg; break;
	case 'g': arg_geometry  = optarg; break;

	case 'i':
	    arg_iterations = strtol (optarg, (char *)0, 0);
	    if (arg_iterations <= 0) {
		usage ("-i argument must be positive.");
	    }
	    break;

	case 'm':
	    arg_mud_max = atof (optarg);
	    if (arg_mud_max < 0.0 || arg_mud_max > 1.0) {
		usage ("-m argument must be in 0.0 .. 1.0.");
	    }
	    break;

	case 'n':
	    arg_mud_num = strtol (optarg, (char *)0, 0);
	    if (arg_mud_num <= 0) {
		usage ("-n argument must be positive.");
	    }
	    break;

	case 'p':
	    arg_pix_size = atof (optarg);
	    if (arg_pix_size <= 0.0) {
		usage ("-p argument must be positive.");
	    }
	    break;

	case 'r': arg_use_root = 1; break;

	case 's':
	    arg_show_time = strtol (optarg, (char *)0, 0);
	    if (arg_show_time <= 0) {
		usage ("-s argument must be positive.");
	    }
            break;

	case 'x': arg_x_center = atof (optarg); break;
	case 'y': arg_y_center = atof (optarg); break;

	default:
	    {
		char buf[80];
		(void) sprintf (buf, "Unrecognized argument '%s'.", optarg);
		usage (buf);
	    }
        }
    }

    if (optind != argc) {
	usage ("Unrecognized argument '%s'.", argv[optind]);
    }
}

/*
 * Given the integer coordinates of a pixel location, make that pixel
 * the center of our image.
 */

void
win_new_center (winp, ix, iy)
win_info_t *winp;
int ix, iy;
{
    winp->x += winp->pix_size * 0.5 * (2 * ix - winp->hpixels);
    winp->y += winp->pix_size * 0.5 * (2 * iy - winp->vpixels);
}

/*
 * Here, the client has just finished generating the image for the
 * root window.  Display the picture and choose new co-ordinates.
 * To choose new co-ordinates, we select two locations at random,
 * such that one location is in the set and one location is out
 * of the set.  These locations become our new coordinates.  The
 * new co-ordinates are selected from within the current co-ordinates.
 * After zooming in for awhile, we begin zooming out.
 */

#define MAX_ROOT_ZOOM_DEPTH 16

void
win_new_root (winp)
win_info_t *winp;
{
    static int depth = 0; /* how far in we are */
    static int dir = 1;

    XSetWindowBackgroundPixmap (winp->dsp, winp->win, winp->pix.p);
    XClearWindow (winp->dsp, winp->win);
    XFreePixmap (winp->dsp, winp->pix.p);

    winp->pix.p = XCreatePixmap (winp->dsp, winp->win,
			 winp->hpixels, winp->vpixels,
			 DefaultDepth (winp->dsp, DefaultScreen (winp->dsp)));

    XSync (winp->dsp, False);
    sleep (1);

    if (dir == -1 ) {
	/*
	 * Zoom out.  Leave the center where it is, but make the pixels
	 * bigger.
	 */
	winp->pix_size *= 3;

	depth -= 1;
	if (depth == 0) dir = 1;
    }
    else {
	/*
	 * Zoom in on a desirable point.  Make the desirable point the
	 * center of the image, and make the pixels smaller.
	 */
	int ix, iy;

	/* get integer coords of pixel */
	ix = (winp->black_list[0].x + winp->color_list[0].x) / 2;
	iy = (winp->black_list[0].y + winp->color_list[0].y) / 2;

	win_new_center (winp, ix, iy);
	winp->pix_size /= 3;

	depth += 1;
        if (depth == MAX_ROOT_ZOOM_DEPTH) dir = -1;
    }
}

int
main (argc, argv)
int argc;
char **argv;
{
	win_info_t *win_create();
	void win_set_lr();

	win_info_t *winp;
	char *client_ctx;
	int resume;

	parse_args (argc, argv);
	winp = win_create ();

	winp->x = arg_x_center;
	winp->y = arg_y_center;
	winp->pix_size = arg_pix_size;

	client_ctx = client_init (winp);

	resume = 0;
	for (;;) {
		/* Let the client run. */
		if (resume) {
			resume = client_resume (client_ctx, winp);
		}
		else {
		        /*
			 * Here, user has asked for a new window.
			 */
			XClearWindow (winp->dsp, winp->win);
			(void) printf ("%s: window at (%g,%g) pix_size=%g\n",
				       argv[0], winp->x, winp->y,
				       winp->pix_size);
			(void) fflush (stdout);

			resume = client_reset (client_ctx, winp);
		}

		if (!arg_use_root) { /* interact with user */
		    resume = win_handle_user (winp, resume);
		}
		else if (!resume) { /* client is done with window */
		    win_new_root (winp);
		}
		else sleep (0); /* don't hog the processor */
	}
}

/*
 * Given a window, change the pixel width and height of the window.
 * We are given a desired width and height.  If possible, we will
 * use this width and height.  However, if the width and height
 * would cause the window to fall off the edge of the screen, then
 * we will reduce the width and height proportionately.
 *
 * Return 1 if we change the width/height; 0 the width/height stay the
 * same; and -1 if the window would become real small.
 *
 * All resizing requests should probably be routed through this
 * routine.
 */

int
win_set_wh (winp, w, h)
win_info_t *winp;
int w, h; /* target width and height */
{
	int scr;
	int dw, dh; /* display width and height */
	int new_x, new_y;
	XWindowAttributes win_attr;
	XSizeHints hints;

	scr = DefaultScreen (winp->dsp);

	dw = DisplayWidth (winp->dsp, scr);
	dh = DisplayHeight (winp->dsp, scr);

	if (w > dw) {
		h = (h * dw) / w; /* preserve shape of window */
		w = dw;
	}
	if (h > dh) {
		w = (w * dh) / h;
		h = dh;
	}

	if (w == winp->hpixels && h == winp->vpixels) return 0;
	if (w < 4 || h < 4) return -1; /* don't get too small */

	winp->hpixels = w;
	winp->vpixels = h;

	/* Adjust the window's size.  If necessary, move the window. */
	XResizeWindow (winp->dsp, winp->win, winp->hpixels, winp->vpixels);
	XGetWindowAttributes (winp->dsp, winp->win, &win_attr);

	new_x = dw - winp->hpixels;
	new_y = dh - winp->vpixels;
	if (win_attr.x > new_x || win_attr.y > new_y) {
		if (win_attr.x < new_x) new_x = win_attr.x;
		if (win_attr.y < new_y) new_y = win_attr.y;

		XMoveWindow (winp->dsp, winp->win, new_x, new_y);

		win_attr.x = new_x;
		win_attr.y = new_y;
	}

	/*
	 * Tell the window manager what we did.  Note:  problems
	 * may or may not occur if the window manager refuses our
	 * request.
	 */
	hints.x = win_attr.x;
	hints.y = win_attr.y;
	hints.width = winp->hpixels;
	hints.height = winp->vpixels;
	hints.flags = USPosition | USSize; /* the user specified these */
	XSetNormalHints (winp->dsp, winp->win, &hints);

	/*
	 * Now, wait for the configuration notify event.  Toss all
	 * other events in the meantime.
	 */
	for (;;) {
		XEvent event;

		XNextEvent (winp->dsp, &event);

		if (event.type == ConfigureNotify
			&& event.xconfigure.window == winp->win)
		{
			/* set width and height to actual values */
			winp->hpixels = event.xconfigure.width;
			winp->vpixels = event.xconfigure.height;
			break;
		}
	}

	return 1;
}

/*
 * Given a window and co-ordinates of a box within that window,
 * adjust the X and Y co-ordinates of the window, and resize the
 * window to the new shape.  Note that the box may actually overlap
 * the window.
 *
 * Return TRUE if we actually change something.  Don't change anything
 * if the new co-ordinates would require either a zero width or a
 * zero height.
 */

int
win_set_box (winp, x1, y1, x2, y2)
win_info_t *winp;
int x1, y1, x2, y2; /* pixel locations of the box */
{
    int ix, iy; /* center pixel of box */
    double old_x, old_y;
    double new_area, old_area;
    double scale;
    int t;

    if ((x1 == x2) || (y1 == y2)) return FALSE;

    /* get length of sides */
    ix = x1 - x2;
    if (ix < 0) ix = -ix;

    iy = y1 - y2;
    if (iy < 0) iy = -iy;

    /* Change the shape of the window.  Preserve the area. */
    old_area = winp->hpixels * winp->vpixels;
    new_area = ix * iy;
    scale = old_area / new_area;
    scale = sqrt (scale);

    old_x = winp->x;
    old_y = winp->y;
    win_new_center (winp, (x1 + x2) / 2, (y1 + y2) / 2);

    t = win_set_wh (winp, (int)(ix * scale), (int)(iy * scale));
    if (t == -1) {
	winp->x = old_x; /* restore old center */
	winp->y = old_y;
	return FALSE;
    }

    winp->pix_size /= scale;

    return TRUE;
}

/*
 * Handle user requests.  If 'resume' is TRUE, we need to
 * return as quickly as possible.  If 'resume' is false,
 * we return when we have something more to compute.
 *
 * Return TRUE if the client can resume drawing in the window.
 * If we modify the window, the client cannot resume.
 */

int
win_handle_user (winp, resume)
win_info_t *winp;
int resume;
{
	int win_new_box ();

	XEvent event;
	char key;
	int i;

	for (;;) {
		/* Get an event. */
		if (resume) {
			if (!XCheckWindowEvent (winp->dsp, winp->win,
				KeyPressMask | ButtonPressMask
				| ButtonReleaseMask | ButtonMotionMask
				| PointerMotionHintMask
				| StructureNotifyMask,
				&event))
			{
				return TRUE;
			}
		}
		else {
			XNextEvent (winp->dsp, &event);
		}

		switch (event.type) {
		case ConfigureNotify:
			/*
			 * See if the window has been resized.
			 */
			if (event.xconfigure.window == winp->win
				&& (event.xconfigure.width != winp->hpixels
				    || event.xconfigure.height != winp->vpixels))
			{
				/* use the width/height we've been given */
				winp->hpixels = event.xconfigure.width;
				winp->vpixels = event.xconfigure.height;
				return FALSE; /* can't resume */
			}
			break;

		case ButtonPress:
			/*
			 * Get a new window.
			 */
			if (win_new_box (winp, event.xbutton.x,
						event.xbutton.y))
			{
				return FALSE; /* can't resume */
			}
			break;

		case KeyPress:
			XLookupString (&event, &key, 1, NULL, NULL);
			switch (key) {
			case 'q': /* quit */
			    exit (0);
			    /* NOTREACHED */

			case 'z':
			    winp->pix_size *= 3;
			    return FALSE;

			case 'e': /* expand window */ 
				i = win_set_wh (winp, winp->hpixels * 2,
						winp->vpixels * 2);

				if (i > 0) return FALSE; /* can't resume */
				break;

			case 's': /* shrink window */
				i = win_set_wh (winp, winp->hpixels / 2,
						winp->vpixels / 2);
				if (i > 0) return FALSE; /* can't resume */
				break;
			}
		}
	}
}

/*
 * The mouse button is currently depressed.  We follow the pointer
 * around drawing a box.  When the mouse is no longer depressed,
 * we attempt to update the window based on the shape and co-ordinates
 * of the new box.
 *
 * Return TRUE if we modify the window, otherwise FALSE.
 */

int
win_new_box (winp, x1, y1)
win_info_t *winp;
int x1, y1; /* starting corner of box */
{
	XEvent event;
	int x2, y2; /* second corner of box */
	int first_time = 1;

	XSetFunction (winp->dsp, winp->gc, GXxor);
	XSetForeground (winp->dsp, winp->gc, -1);
	XSetBackground (winp->dsp, winp->gc, -1);

	for (;;) {
		XNextEvent (winp->dsp, &event);
		switch (event.type) {
		case ButtonRelease:
			/*
			 * We have selected a new window.  Reset the
			 * window and save the co-ordinates.
			 */
			/* erase current box */
			if (!first_time) {
				XDrawRectangle (winp->dsp, winp->win,
					winp->gc, x1, y1, x2-x1, y2-y1);
			}

			x2 = event.xbutton.x;
			y2 = event.xbutton.y;

			/* reset drawing mode */
			XSetFunction (winp->dsp, winp->gc, GXcopy);

			/* Compute new co-ordinates and resize window. */
			return win_set_box (winp, x1, y1, x2, y2);

		case MotionNotify: {
			/*
			 * Redraw our box.
			 */

			Window root, child;
			int root_x, root_y, win_x, win_y;
			unsigned int keys_buttons;

			while (XCheckMaskEvent (winp->dsp, ButtonMotionMask,
					&event));

			if (!XQueryPointer (winp->dsp, event.xmotion.window,
				&root, &child, &root_x, &root_y,
				&win_x, &win_y, &keys_buttons))
			{
				break; /* not in this window */
			}

			/* undraw the previous box, if it exists */
			if (!first_time) {
				XDrawRectangle (winp->dsp, winp->win,
					winp->gc, x1, y1, x2-x1, y2-y1);
			}

			x2 = win_x;
			y2 = win_y;

			/* draw new box */
			first_time = 0;
			XDrawRectangle (winp->dsp, winp->win,
				winp->gc, x1, y1, x2-x1, y2-y1);
			break;
		    }
		}
	}
}

/*
 * Create the initial window.
 */

win_info_t *
win_create ()
{
	Colormap init_color_map ();
	Colormap read_color_map ();
	void win_display_colormap ();

	win_info_t *winp;
	int scr;
	Colormap cmap;
	XSetWindowAttributes xswa;
	XColor black, white;
	XWMHints wm_hints;
	int winX, winY, winH, winW;
    	int cursor_shape1 = XC_watch;
    	int cursor_shape2 = XC_top_left_arrow;
	int bufsize;
	int i;

	winp = (win_info_t *) malloc (sizeof (*winp));
	if (!winp) {
		(void) fprintf (stderr, "win_create: Out of memory.\n");
		exit (1);
	}

	winp->dsp = XOpenDisplay (arg_display);
	if (!winp->dsp) {
		perror ("win_create: Cannot open display.");
		exit (1);
	}

	scr = DefaultScreen (winp->dsp);

	if (arg_colorfile) {
		cmap = read_color_map (winp);
		if (winp->color_len <= 1) {
			if (winp->pixels) free ((char *)winp->pixels);
			cmap = init_color_map (winp);
		}
	}
	else cmap = init_color_map (winp);

	black.pixel = BlackPixel (winp->dsp, scr);
	if (!arg_use_root) XQueryColor (winp->dsp, cmap, &black);
	winp->black = black.pixel;

	white.pixel = WhitePixel (winp->dsp, scr);
	if (!arg_use_root) XQueryColor (winp->dsp, cmap, &white);
	winp->white = white.pixel;

	if (arg_use_root) { /* let's be a background pattern... */
		winp->win = DefaultRootWindow (winp->dsp);
		winX = 0;
		winY = 0;
		winW = DisplayWidth  (winp->dsp, scr);
		winH = DisplayHeight (winp->dsp, scr);
		winp->gc = DefaultGC (winp->dsp, scr);

		/*
		 * Buffer the image in a pixmap.
		 */
		winp->pix.p = XCreatePixmap (winp->dsp, winp->win,
					     winW, winH,
					     DefaultDepth (winp->dsp, scr));
	}
	else {
		winW = HPIXELS;
		winH = VPIXELS;
		winX = 0;
		winY = 0;

		if (arg_geometry) {
			XParseGeometry (arg_geometry, &winX, &winY,
							&winW, &winH);
		}

		xswa.event_mask = 0;
		xswa.background_pixel = black.pixel;
		xswa.backing_store = WhenMapped;

		winp->win = XCreateWindow (winp->dsp,
			RootWindow (winp->dsp, scr),
			winX, winY, winW, winH, 0,
			DefaultDepth (winp->dsp, scr), InputOutput,
			DefaultVisual (winp->dsp, scr),
			CWEventMask | CWBackPixel | CWBackingStore,
			&xswa);

		/* Tell the window manager about ourself. */
		XStoreName (winp->dsp, winp->win, "Mandel");
		XSetIconName (winp->dsp, winp->win, "Mandel");

		wm_hints.input = True; /* passive input */
		wm_hints.initial_state = NormalState;
		wm_hints.flags = InputHint | StateHint;
		XSetWMHints (winp->dsp, winp->win, &wm_hints);

		XMapWindow (winp->dsp, winp->win);
		XSync (winp->dsp, False);
	
		XSelectInput (winp->dsp, winp->win,
			KeyPressMask | ButtonPressMask | ButtonReleaseMask
			| StructureNotifyMask | ButtonMotionMask
			| PointerMotionHintMask);

		winp->gc = XCreateGC (winp->dsp, winp->win, 0, NULL);

		/*
		 * Render the image directly to the screen.
		 */
		winp->pix.w = winp->win;

		winp->busy = XCreateFontCursor (winp->dsp, cursor_shape1);
		winp->nbusy = XCreateFontCursor (winp->dsp, cursor_shape2);
		XDefineCursor (winp->dsp, winp->win, winp->nbusy);
	}

	winp->hpixels = winW;
	winp->vpixels = winH;

	winp->black_num = winp->color_num = 0; /* empty */
	for (i = 0; i < SAVE_LIST_LEN; i++) {
	    winp->black_list[i].x =
	    winp->black_list[i].y =
	    winp->color_list[i].x =
	    winp->color_list[i].y = 0;
	}

	XSetForeground (winp->dsp, winp->gc, white.pixel);
	XSetBackground (winp->dsp, winp->gc, black.pixel);
	XSetFunction (winp->dsp, winp->gc, GXcopy);

	/*
	 * Having problems on the Sun.  Put in a clear and sync
	 * in an attempt to delay some messages.
	 */
	sleep (1);
	XClearWindow (winp->dsp, winp->win);
	XSync (winp->dsp, False);

	if (arg_show_time) win_display_colormap (winp);

	/*
	 * Allocate a buffer area for imaging.  The purpose of this
	 * buffer is to reduce the amount of work required from the
	 * XServer.  We allocate a buffer one line high and the
	 * full width of the screen.
	 *
	 * Note:  We use an XYPixmap because the Apollo workstations
	 * seem to have problems with ZPixmaps.
	 */
#define USE_ZPIXMAP
	winp->image = XCreateImage (winp->dsp, 
			DefaultVisual (winp->dsp, scr),
			DefaultDepth (winp->dsp, scr),
#ifdef USE_ZPIXMAP
			ZPixmap,
#else
			XYPixmap,
#endif
			0, /* offset to first pixel */
			(char *) 0, /* pointer to image data */
			DisplayWidth  (winp->dsp, scr),
			1, /* height of image */
			32, /* alignment of scan lines */
			0);
	if (!winp->image) {
		(void) fprintf (stderr,
			"win_create: Cannot allocate image.\n");
		exit (1);
	}
#if 1
	winp->image->byte_order = 0;
	winp->image->bitmap_bit_order = 0;
#endif

	/* Allocate a data area for the image. */
#ifdef USE_ZPIXMAP
	/* use this calculation for ZPixmap format */
	bufsize = winp->image->bytes_per_line;
#else
	/* use this calculation for XYPixmap format */
	bufsize = winp->image->bytes_per_line * winp->image->depth;
#endif
	winp->image->data = malloc ((unsigned) bufsize);
	if (!winp->image->data) {
		(void) fprintf (stderr,
		  "win_create: Cannot allocate %d bytes for image data.\n",
		  bufsize);
		exit (1);
	}

	return winp;
}

/*
 * Display the initial colormap.  We can handle up to 256 colors here.
 */

void
win_display_colormap (winp)
win_info_t *winp;
{
	int i;
        int r, c;
        int pixel;
 
        for (i = 0; i < winp->color_len; i++) {
                if (winp->pixels) pixel = winp->pixels[i];
		else pixel = i;

		c = i % 16;
		r = i / 16;
 
                XSetForeground (winp->dsp, winp->gc, pixel);
                XFillRectangle (winp->dsp, winp->win, winp->gc,
                        (winp->hpixels * c) / 16,
                        (winp->vpixels * r) / 16,
                        winp->hpixels / 16,
                        winp->vpixels / 16);
#if 0
		/*
		 * On the 'sun', this doesn't seem to work unless
		 * we have this 'XSync' lying around.
		 */
		XSync (winp->dsp, False);
		sleep (1);
#endif
        }

	XSync (winp->dsp, False);
	sleep (arg_show_time);
}

/*
 * Read colors for the colormap from a file.  Return the colormap.
 */

#define COLORLEN 80
#define is_white(c) ((c) == ' ' || (c) == '\t')
#define scan_white(p) while (*(p) && is_white(*(p))) (p) += 1
#define scan_black(p) while (*(p) && !is_white(*(p))) (p) += 1

static Colormap
read_color_map (winp)
win_info_t *winp;
{
	FILE *f;
	char colorname[COLORLEN];
	int scr;
	int npixels;
	int i;
	Colormap cmap;
	XColor xcolor;

	scr = DefaultScreen (winp->dsp);

	/* set initial length and offset */
	winp->color_len = XDisplayCells (winp->dsp, scr);
	winp->pixels = NULL;

	cmap = XDefaultColormap (winp->dsp, scr);

	/* open the file of color names */
	f = fopen (arg_colorfile, "r");
	if (!f) {
		(void) fprintf (stderr,
			"read_color_map: Cannot open file %s.\n",
			arg_colorfile);
		perror ("");
		return cmap;
	}

	/* allocate pixel storage */
	for (npixels = winp->color_len; npixels; npixels--) {
		winp->pixels = (unsigned long *) malloc (
				sizeof (*(winp->pixels)) * npixels);
		if (winp->pixels) break;
	}

	if (!npixels) {
		(void) fprintf (stderr,
			"read_color_map: Can't allocate pixel storage.\n");
		winp->color_len = 0;
		return cmap;
	}

	/* read the colors */
	for (i = 0; i < npixels; ) {
		char *p, *start;

		if (fgets (colorname, COLORLEN, f) == NULL) break;
		p = strchr (colorname, ';');
		if (!p) p = strchr (colorname, '\n');
		if (p) *p = '\0'; /* toss end of line */

		start = colorname;
		scan_white (start); /* find start of string */

		p = start;
		scan_black (p); /* find end of string */
		if (p == start) continue; /* empty line */

		*p = '\0'; /* terminate the string */

		if (!XParseColor (winp->dsp, cmap, start, &xcolor)) {
			(void) fprintf (stderr,
			"read_color_map: color name '%s' isn't in database.\n",
				start);
			continue;
		}

		if (!XAllocColor (winp->dsp, cmap, &xcolor)) {
			/*
			 * Although we can't allocate any more colors,
			 * we may be able to share some already existing
			 * color.  Keep looping to give it a try.
			 */
			continue;
		}

		winp->pixels[i] = xcolor.pixel;
		i += 1;
	}

	winp->color_len = i; /* remember length of pixel array */

	(void) fclose (f);
	return cmap;
}

/*
 * Allocate a color map.
 */

static Colormap
init_color_map (winp)
win_info_t *winp;
{
	void convertHSI_to_RGB();

	int scr;
	Colormap cmap;
	double hue, intensity, saturation;
	int r, g, b;
	XColor celldef;
	int i;
	int npixels;
	double mud;
	int m;
	int dir;

	scr = DefaultScreen (winp->dsp);

	/* set initial length and offset */
	winp->color_len = XDisplayCells (winp->dsp, scr);
	winp->pixels = NULL;

	cmap = XDefaultColormap (winp->dsp, scr);

	/* allocate color map */
	for (npixels = winp->color_len; npixels; npixels--) {
		winp->pixels = (unsigned long *) malloc (
				sizeof (*(winp->pixels)) * npixels);
		if (!winp->pixels) continue;

		if (!XAllocColorCells (winp->dsp, cmap, False,
			(unsigned long *)0, 0,
			winp->pixels, npixels))
		{
			free ((char *)winp->pixels);
			continue;
		}

		break;
	}

	if (!npixels) {
		(void) fprintf (stderr,
			"init_color_map: Can't allocate colormap.\n");
		return cmap;
	}

	/* Set the colors for each allocated pixel. */

	winp->color_len = npixels;
	hue = 0.0;
	mud = 0.0;
	dir = 1;

	for (i = 0; i < npixels; i++) {
		convertHSI_to_RGB (hue, mud, &r, &g, &b);
		celldef.pixel = winp->pixels[i];
		celldef.red = r;
		celldef.green = g;
		celldef.blue = b;
		celldef.flags = DoBlue | DoGreen | DoRed;
		celldef.pad = NULL;

		XStoreColor (winp->dsp, cmap, &celldef);

		m = i % arg_mud_num;
		if (m == arg_mud_num - 1) { /* increment hue */
			hue += 1.0 * arg_mud_num / (double) (npixels);
			dir = -dir; /* change direction */
		}
		else if (dir > 0) { /* increase mud */
			mud = mud + arg_mud_max / arg_mud_num;
		}
		else { /* decrease mud */
			mud = mud - arg_mud_max / arg_mud_num;
		}
	}

	return cmap;
}

/*
 * Given a hue, saturation, and brightness (intensity), generate
 * r, g, and b values.  Here, we ignore the saturation and brightness
 * since we know what we want.  We use sine curves to ramp values up
 * and down quickly, so that the colors will be more interesting.
 */

#define ramp_up(h) sin (M_PI_2 * (h))
#define ramp_down(h) cos (M_PI_2 * (h))

void
convertHSI_to_RGB (hue, mud, R, G, B)
double hue, mud;
int *R,*G,*B ;

{
 double r,g,b ;
 int i;

 hue = (double) (hue*6) ;
 i = (int) (hue);
 hue = hue - i; /* normalize 'hue' into range 0..1 */
 switch (i)
 {
  case 0 :
        r = 1.0;
        g = ramp_up (hue);
        b = ramp_up (mud);
        break;
  case 1:
        r = ramp_down (hue);
        g = 1.0;
        b = ramp_up (mud);
        break;
  case 2:
        r = ramp_up (mud);
        g = 1.0;
        b = ramp_up (hue);
        break;
  case 3:
        r = ramp_up (mud);
        g = ramp_down (hue);
        b = 1.0;
        break;
  case 4:
        r = ramp_up (hue);
        g = ramp_up (mud);
        b = 1.0;
        break;
  case 5:
        r = 1.0;
        g = ramp_up (mud);
        b = ramp_down (hue);
        break;
  }

 *R = (int) (r*65535) ;
 *G = (int) (g*65535) ;
 *B = (int) (b*65535) ;

}


/*
 * Flush the graphics package.
 */

void
win_flush (winp)
win_info_t *winp;
{
	XSync (winp->dsp, False);
}

/*
 * Change the shape of the cursor on the screen.  Only done if not
 * on root.
 */

void
win_busy (winp, state)
win_info_t *winp;
int state;
{
    if (!arg_use_root) {
	XDefineCursor (winp->dsp, winp->win, state ? winp->busy : winp->nbusy);
    }
}

/*
 * Draw a colored line.  The 'color' is a postive number.  We
 * will take this color modulo the available colors.
 */

void
win_draw_colored_line (winp, color, from_x, from_y, to_x, to_y)
win_info_t *winp;
int color;
int from_x, from_y, to_x, to_y;
{
	int pixel;

	pixel = color % winp->color_len;
	if (winp->pixels) pixel = winp->pixels[pixel];

	XSetForeground (winp->dsp, winp->gc, pixel);
	XDrawLine (winp->dsp, winp->win, winp->gc, from_x, from_y, to_x, to_y);
}

/*
 * Given a run-length encoded line, draw the line on the screen.
 * Here, we buffer the line into an image, and then send the image
 * to the screen.  This technique is designed to reduce the amount
 * of work required by the XServer.
 *
 * If we understood the internal format of an Image better, we could
 * probably generate the image data using low-level techniques that
 * would greatly enhance the speed of this routine.  Eventually, we
 * might want to farm out some of these computations to the compute
 * servers, but that would run the risk of making the compute servers
 * more difficult to port.
 *
 * When we are imaging on the root window, we store the pixel
 * co-ordinates where we think it would be interesting to expand
 * the image.  We willl choose the right side of a black line
 * that is followed by a colored line.  And we will prefer a short
 * line.
 */

void
win_render_line (winp, buf, buflen, linenum)
win_info_t *winp;
color_x_t *buf; /* run-length encoded line */
int buflen; /* number of elements pointed to by 'buf' */
int linenum; /* line number of the window to be drawn */
{
	int last_x;
	int i, j;
	unsigned long pixel;

	/* Fill the image with colors. */
	last_x = 0;
	for (i = 0; i < buflen; i++) {
		if (buf[i].color == arg_iterations) { /* black? */
			pixel = winp->black;

			/* record various black and colored locations */
			if (arg_use_root) {
			    if (i < buflen - 1
				&& buf[i].x - last_x + 1 < 5)
			    {
				winp->black_list[0].x = buf[i].x + 1;
				winp->black_list[0].y = linenum;
				winp->color_list[0].x = buf[i].x + 1;
				winp->color_list[0].y = linenum;
			    }
			}
		}
		else {
			pixel = buf[i].color % winp->color_len;
			if (winp->pixels) pixel = winp->pixels[pixel];
		}

		/* Fill a line segment with this pixel. */
		for (j = last_x; j <= buf[i].x; j++) {
			XPutPixel (winp->image, j, 0, pixel);
		}

		last_x = buf[i].x + 1;
	}

	/* Put the image on the screen. */
	XPutImage (winp->dsp, winp->pix.w, winp->gc, winp->image,
			0, 0, 0, linenum, winp->hpixels, 1);
}

/*
 * Draw the image that we last drew on the screen.  This exists
 * purely to take advantage of the symmetry that exists about the
 * X-axis of the mandelbrot set.  This routine isn't absolutely
 * necessary.
 */

void
win_rerender_line (winp, linenum)
win_info_t *winp;
int linenum;
{
	XPutImage (winp->dsp, winp->pix.w, winp->gc, winp->image,
			0, 0, 0, linenum, winp->hpixels, 1);
}
