/*
 * Mandel by Rob Swiston and Chuck Simmons
 * Copyright (C) 1989, 1990 by Rob Swiston and Chuck Simmons
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 1 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * This file handles the code to implement a remote procedure call
 * for the mandelbrot set.  There are two portions to this package:
 * the portion linked into the client side (the side initiating the
 * rpc), and the portion linked into the server side (the side implementing
 * the procedure call).
 *
 * This 'rpc' is a bit wierd in that on the client side, the 'rpc'
 * executes asynchronously.  That is, the client side will call a
 * routine to initiate the rpc, passing arguments to the remote procedure
 * and passing a special value that identifies this call.  Later, the
 * client will notice that the 'rpc' has finished and will call another
 * routine to pick up the results of the call.
 *
 * Since the special identifer value used on the client side is used
 * purely to identify a packet, this identifier is not passed to the
 * subroutine on the server side.
 *
 * We handle all necessary byte swapping in this routine.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include "mandel.h"

extern int errno; /* not defined in errno.h on pyramid */

/*
 * Routines to perform floating point conversion.  Our standard format
 * is that used by the MIPS R2000, Sun, and Sequent systems.
 */

#ifdef VAX_BYTE_ORDER
/*
 * Swap bytes for IEEE floating point.
 */
double
htond (x)
double x;
{
	char *p, *q;
	double y;
	int i;

	p = (char *) &x;
	q = (char *) &y;

	for (i = 0; i < sizeof (double); i++) {
		q[sizeof (double) - i - 1] = p[i];
	}

	return y;
}

# define ntohd(x) htond (x)
# define double_done
#endif

#ifdef CONVEX
/*
 * Convert between Convex floating point format and IEEE floating point.
 * We need to subtract two from the exponent of the Convex format.
 *
 * These routines aren't real robust for very small or very large numbers.
 */

double
htond (x)
double x;
{
	long *l1, *l2;
	double y;

	l1 = (long *) &x;
	l2 = (long *) &y;

	y = x; /* init result */
	l2[0] = l1[0] - 0x200000; /* fix up exponent */
	return y;
}

/* Add two going the other way. */

double
ntohd (x)
double x;
{
	long *l1, *l2;
	double y;

	l1 = (long *) &x;
	l2 = (long *) &y;

	y = x; /* init result */
	l2[0] = l1[0] + 0x200000; /* fix up exponent */
	return y;
}

# define double_done
#endif

#ifdef IS_370
/*
 * Convert between IBM 370 floating point format and IEEE floating point
 * format.
 */

double
htond (x)
double x;
{
	int exp;
	int sign;
	long long frac;
	double y;
	long *l;
	long long *ll;

	ll = (long long *) &x;
	frac = *ll; /* fraction and other stuff */
	exp = (frac >> 56) & 0x7f; /* extract exponent */
	sign = (frac >> 32) & 0x80000000; /* extract sign */

	/* build IEEE format */
	frac = (frac >> 4) & 0xfffffffffffff;

	/*
	 * Convert the exponent.  In IBM land, this is a signed
	 * exponent used as a power of 16.  We want a power of
	 * 2, so we make the exponent signed, multiply by 4,
	 * and then add on the bias of 1023.
	 */
	exp = (exp << 25) >> 25; /* make signed */
	exp *= 4; /* go from base 16 to base 2 */
	exp += 1023; /* add on the bias */

	exp = (exp & 0x7ff) << 20;

	ll = (long long *) &y;
	*ll = frac;

	l = (long *) &y;
	*l |= sign | exp;

(void) printf ("htond: converted (%f) 0x%x 0x%x to 0x%x 0x%x\n",
		%f, ((long *)(&x))[0], ((long *)(&x))[1],
		((long *)(&y))[0], ((long *)(&y))[1]);

	return y;
}

/* Go the other way. */

double
ntohd (x)
double x;
{
	int exp;
	int sign;
	long long frac;
	double y;
	long *l;
	long *ll;
	int i;

	ll = (long long *) &x;
	frac = *ll; /* fraction and other stuff */
	exp = (frac >> 52) & 0x7ff; /* extract exponent */
	sign = (frac >> 32) & 0x80000000; /* extract sign */

	/* build IEEE format */
	frac = frac << 4;

	/*
	 * Convert the exponent.  First we subtract off the IEEE bias.
	 * Then, we need to divide by four.  If exp is not a multiple
	 * of four, then we shift the fraction one to the right and
	 * increment the exponent until the exponent is a multiple of
	 * four.
	 */
	exp -= 1023; /* add on the bias */

	i = 4 - (exp & 3);
	exp += i;
	frac >>= i;

	exp /= 2;
	exp = (exp << 24) & 0x7f000000;

	ll = (long long *) &y;
	*ll = frac & 0xfffffffffffff;

	l = (long *) &y;
	*l |= sign | exp;

(void) printf ("ntohd: converted 0x%x 0x%x to 0x%x 0x%x (%f)\n",
		((long *)(&x))[0], ((long *)(&x))[1],
		((long *)(&y))[0], ((long *)(&y))[1], y);

	return y;
}

# define double_done
#endif

#ifndef double_done
/*
 * Use this for IEEE floating point when VAX byte ordering is not
 * present.
 */
# define htond(x) (x)	/* host-to-network double */
# define ntohd(x) (x)	/* network-to-host double */
#endif

/*
 * Define the structure we use for passing information between client
 * and server.
 */
typedef struct net_parms {
	double x_start, y_start; /* starting coords of line */
	double x_end; /* ending coord of line */
	short npixels; /* number of pixels in line */
	short max_color; /* maximum colors per pixel */
	long unused; /* pad this out to a multiple of 8 bytes */
} net_parms_t;

/*
 * Define the structure used for passing the response.
 */
/* The following doesn't work very well because the pyramid pads this
 * out to be four bytes instead of the desired two bytes:
 *
 *	typedef struct net_response {
 *		short buflen; / * number of elements following this header * /
 *	} net_response_t;
 */
typedef short net_response_t;

#ifdef CLIENT
/*
 * Provide an interface to the remote routine for the client.
 * Return 0 on success, -1 on error.  If an error occurs, 'errno'
 * will be set.
 */

int
net_csend (sock, x_start, y_start, x_end, npixels, max_color)
int sock;
double x_start, y_start;
double x_end;
int npixels, max_color;
{
	net_parms_t parms;
	int len;

	/* package the parameters */
	parms.x_start = htond (x_start);
	parms.y_start = htond (y_start);
	parms.x_end = htond (x_end);
	parms.npixels = htons (npixels);
	parms.max_color = htons (max_color);

	len = write (sock, &parms, sizeof (parms));
	if (len == -1) return -1; /* return any error */

	if (len < sizeof (parms)) { /* partial write? */
		errno = EIO;
		return -1;
	}

	return 0; /* success */
}
#endif

/*
 * Experience shows that when we issue a read, we may not read all
 * of the requested bytes, probably because of packetizing on the
 * network.  Here, we will issue multiple reads, if necessary, to
 * read all desired bytes.
 */

static int
net_read (sock, buf, buflen)
int sock; /* socket to read */
char *buf; /* buffer to fill */
int buflen; /* length to read */
{
	int left;
	int len;
	char *p;

	left = buflen;
	p = buf;
	while (left > 0) {
		len = read (sock, p, left);
		if (len == -1) return -1; /* error occured */
		if (len == 0) break; /* can't read any more */

		left -= len;
		p += len;
	}

	return buflen - left;
}

#ifdef CLIENT
/*
 * Provide a routine to return values to the client.  The client gives
 * us an array that we fill, and we return a length to the client.
 */

int
net_crecv (sock, buf)
int sock; /* socket to read */
color_x_t *buf; /* buffer to fill */
{
	net_response_t response;
	int len;
	int buflen;

	/* read the response header */
	len = net_read (sock, (char *) &response, sizeof (response));
	if (len == -1) return -1; /* oops */
	if (len < sizeof (response)) {
		(void) fprintf (stderr,
			"net_crecv: reading header received %d of %d bytes\n",
			len, sizeof (response));
		errno = EIO;
		return -1;
	}

	/* read buffer */
	buflen = ntohs (response);
	len = net_read (sock, (char *) buf, sizeof (*buf) * buflen);
	if (len == -1) return -1;
	if (len < sizeof (*buf) * buflen) {
		(void) fprintf (stderr,
			"net_crecv: reading buffer received %d of %d bytes\n",
			len, sizeof (*buf) * buflen);
		errno = EIO;
		return -1;
	}

	/* byteswap the buffer contents */
	for (len = 0; len < buflen; len++) {
		buf[len].x = ntohs (buf[len].x);
		buf[len].color = ntohs (buf[len].color);
	}
	return buflen;
}
#endif

/*
 * The server routine.  This routine loops reading requests from the
 * client and sending back responses.
 */

#ifdef SERVER
void
net_server (sock)
int sock;
{
	net_parms_t parms;
	net_response_t *response;
	int len, wlen, i;
	color_x_t *buf;

	for (;;) { /* until an error occurs */
		/* read a request */
		len = net_read (sock, &parms, sizeof (parms));
		if (len == -1) {
			perror ("net_server: cannot read socket");
			exit (1);
		}
		if (len != sizeof (parms)) {
			if (len == 0) exit (0); /* client went away */
			(void) fprintf (stderr,
				"net_server: read %d of %d bytes.\n",
				len, sizeof (parms));
			exit (1);
		}

		/* Byte swap the request. */
		parms.x_start = ntohd (parms.x_start);
		parms.y_start = ntohd (parms.y_start);
		parms.x_end = ntohd (parms.x_end);
		parms.npixels = ntohs (parms.npixels);
		parms.max_color = ntohs (parms.max_color);

		/* Allocate storage for the response. */
		len = sizeof (*response) + parms.npixels * sizeof (color_x_t);
		response = (net_response_t *) malloc (len);
		if (!response) {
			(void) fprintf (stderr, "net_server: Out of memory.\n");
			exit (1);
		}
		buf = (color_x_t *) (response + 1);

		/* Call the remote procedure. */
		len = mandel_line (parms.x_start, parms.y_start, parms.x_end,
					parms.npixels, parms.max_color, buf);

		/* byteswap the buffer contents */
		for (i = 0; i < len; i++) {
			buf[i].x = htons (buf[i].x);
			buf[i].color = htons (buf[i].color);
		}

		/* Send the response to the client. */
		*response = htons (len);

		wlen = sizeof (*response) + len * sizeof (color_x_t);
		len = write (sock, (char *) response, wlen);
		if (len == -1) {
			perror ("net_server: cannot write socket");
			exit (1);
		}
		if (len != wlen) {
			(void) fprintf (stderr,
				"net_server: wrote %d of %d bytes.\n",
				len, wlen);
			exit (1);
		}

		free ((char *) response);
	}
}

/*
 * Server routine to listen for a connection.  In response, we fork
 * and return the created connection.  The parent routine continues
 * to listen.  Return -1 and set errno if we fail at some point.
 */

int
net_slisten (port)
int port; /* port on which to listen */
{
	int sock, new_sock;
	struct sockaddr_in sin, from;
	int len;

	bzero ((char *) &sin, sizeof (sin));
	sin.sin_family = AF_INET;
	sin.sin_port = port;

	sock = socket (AF_INET, SOCK_STREAM, 0);
	if (sock < 0) return -1;

	if (bind (sock, &sin, sizeof (sin)) < 0) return -1;

	listen (sock, 1);
	for (;;) {
		len = sizeof (from);
		new_sock = accept (sock, &from, &len);
		if (new_sock < 0) {
			if (errno == EINTR) continue;
			return -1;
		}

		switch (fork ()) {
		case -1: /* error? */
			return -1;

		case 0: /* child */
			(void) close (sock);
			return new_sock;

		default: /* parent */
			(void) close (new_sock);
			break;
		}
	}
	/* NOTREACHED */
}
#endif			

/*
 * On the client side, open a connection to a server process.  Return
 * the socket for the connection.  If something goes wrong, return -1
 * with 'errno' set.
 */

#ifdef CLIENT
int
net_copen (host, port)
char *host;
int port;
{
	int sock;
	struct sockaddr_in sin;
	struct hostent *hp;

	hp = gethostbyname (host);
	if (!hp) return -1;

	bzero ((char *) &sin, sizeof (sin));
	bcopy (hp->h_addr, (char *) &(sin.sin_addr), hp->h_length);
	sin.sin_family = hp->h_addrtype;
	sin.sin_port = port;

	sock = socket (AF_INET, SOCK_STREAM, 0);
	if (sock < 0) return -1;

	if (connect (sock, (char *) &sin, sizeof (sin)) < 0) return -1;

	return sock;
}
#endif

